//
//  n80config.m
//
//  Copyright (C) 1993-2024 Hiroshi Kanemoto.
//    Released under the MIT license
//    http://opensource.org/licenses/mit-license.php
//
//    Created by hiroshi on 2024/05/15.
//
#import <Foundation/Foundation.h>
#import <N80X-Swift.h>
#include <ctype.h>
#include <string.h>
#import "n80em.h"
#import "n80config.h"
#import "n80keyboard.h"
#import "n80graph.h"
extern N80LX        n80lx;
/**--------------------------------------------------------------------------
 **	declare macro
 **------------------------------------------------------------------------*/
#define TYPE_TOGGLE     (0)     // toggle control
#define TYPE_TEXTIN     (1)     // text-in control
#define TYPE_DRVSELECT  (9)     // drive selector
#define TYPE_IMGSLOT  (100)     // disk image slot
#define TYPE_TERMINATOR (-1)    // terminator dummy control
#define VERB_INIT       (0)     // INIT control
#define VERB_EVENT      (1)     // Key Event handle
#define VERB_PRINT      (2)     // Print control
#define VERB_PRINT_SEL  (3)     // Print control (selected status)
#define VERB_COMMIT		(4)     // on toggled set value
#define VERB_TEXTIN_CH  (5)     // on key char
#define ACT_NONE        (0)
#define ACT_DONE        (1)
#define ACT_CHANGEPAGE  (9)
#define ACT_QUIT        (-1)
#define CHAR_VALID      (1)
#define CHAR_INVALID    (0)
#define SP              (32)
//-----------------------------------------------------------------------------
//  local function declaration
//-----------------------------------------------------------------------------
static int _controller_tenkeyless(LPCFGITEM, int, LPVOID);
static int _controller_interrace(LPCFGITEM, int, LPVOID);
static int _controller_nowait(LPCFGITEM, int, LPVOID);
static int _controller_cmtfile(LPCFGITEM, int, LPVOID);
static int _controller_n80rom(LPCFGITEM, int, LPVOID);
static int _controller_fdurom(LPCFGITEM, int, LPVOID);
static int _controller_image_slot(LPCFGITEM, int, LPVOID);
static int _controller_drive_select(LPCFGITEM, int, LPVOID);
//-----------------------------------------------------------------------------
//	local variables
//-----------------------------------------------------------------------------
static CFGITEM page0_items[] = {
	{0x01, TYPE_TOGGLE, "TEN-Keyless .......................... [ ]", "assign Ten-Key (2/4/6/8) to cursor key for Tenkey-less keyboard", 0, _controller_tenkeyless},
	{0x02, TYPE_TOGGLE, "Interrace-like effect ................ [ ]", "display like interrace", 0, _controller_interrace},
	{0x03, TYPE_TOGGLE, "No-wait .............................. [ ]", "ignore memory wait, M1 wait", 0, _controller_nowait},
	{0x11, TYPE_TEXTIN, "N80 rom file name .................... [                        ]", "N-BASIC ROM file name (required)", 24, _controller_n80rom},
	{0x11, TYPE_TEXTIN, "Disk Unit rom file name .............. [                        ]", "Disk ROM file name (optional)", 24, _controller_fdurom},
	{0x31, TYPE_TEXTIN, "output cmt file name ................. [                 ]", "cmt output file name when csave or mon W command (required)", 17-1, _controller_cmtfile},
	{0x41, TYPE_IMGSLOT, "Image File Slot #0 Control ........... [ ]", "[DEL] remove [INS] load [RET] insert/eject disk image", 0, _controller_image_slot},
	{0x42, TYPE_IMGSLOT, "Image File Slot #1 Control ........... [ ]", "[DEL] remove [INS] load [RET] insert/eject disk image", 0, _controller_image_slot},
	{0x00, TYPE_TERMINATOR}
};
static CFGITEM page1_items[] = {
	{0x01, TYPE_DRVSELECT, " 0) .................................. [ ]", "inser/eject disk image #0 to/from drive", -1, _controller_drive_select},
	{0x02, TYPE_DRVSELECT, " 1) .................................. [ ]", "inser/eject disk image #1 to/from drive", -1, _controller_drive_select},
	{0x03, TYPE_DRVSELECT, " 2) .................................. [ ]", "inser/eject disk image #2 to/from drive", -1, _controller_drive_select},
	{0x04, TYPE_DRVSELECT, " 3) .................................. [ ]", "inser/eject disk image #3 to/from drive", -1, _controller_drive_select},
	{0x05, TYPE_DRVSELECT, " 4) .................................. [ ]", "inser/eject disk image #4 to/from drive", -1, _controller_drive_select},
	{0x06, TYPE_DRVSELECT, " 5) .................................. [ ]", "inser/eject disk image #5 to/from drive", -1, _controller_drive_select},
	{0x07, TYPE_DRVSELECT, " 6) .................................. [ ]", "inser/eject disk image #6 to/from drive", -1, _controller_drive_select},
	{0x08, TYPE_DRVSELECT, " 7) .................................. [ ]", "inser/eject disk image #7 to/from drive", -1, _controller_drive_select},
	{0x09, TYPE_DRVSELECT, " 8) .................................. [ ]", "inser/eject disk image #8 to/from drive", -1, _controller_drive_select},
	{0x0A, TYPE_DRVSELECT, " 9) .................................. [ ]", "inser/eject disk image #9 to/from drive", -1, _controller_drive_select},
};
static LPCFGITEM page_items[] = { page0_items, page1_items };
//-----------------------------------------------------------------------------
//	local functions
//-----------------------------------------------------------------------------
static char to_ascii(LPKEYEVT kv, GCKeyCode gckey) {
	kv->key = 0x0;
    if (gckey == GCKeyCodeOne)          /* 1 */	return kv->SHIFT ? 0x21 : 0x31;
    else if (gckey == GCKeyCodeTwo)     /* 2 */	return kv->SHIFT ? 0x22 : 0x32;
    else if (gckey == GCKeyCodeThree)	/* 3 */	return kv->SHIFT ? 0x23 : 0x33;
    else if (gckey == GCKeyCodeFour)    /* 4 */	return kv->SHIFT ? 0x24 : 0x34;
    else if (gckey == GCKeyCodeFive)    /* 5 */	return kv->SHIFT ? 0x24 : 0x35;
    else if (gckey == GCKeyCodeSix)     /* 6 */	return kv->SHIFT ? 0x26 : 0x36;
    else if (gckey == GCKeyCodeSeven)	/* 7 */	return kv->SHIFT ? 0x27 : 0x37;
    else if (gckey == GCKeyCodeEight)	/* 8 */	return kv->SHIFT ? 0x28 : 0x38;
    else if (gckey == GCKeyCodeNine)    /* 9 */	return kv->SHIFT ? 0x29 : 0x39;
    else if (gckey == GCKeyCodeZero)    /* 0 */	return kv->SHIFT ? 0x30 : 0x30;

    else if (gckey == GCKeyCodeKeyQ)	/* Q */	return kv->SHIFT ? 0x51 : 0x71;
    else if (gckey == GCKeyCodeKeyW)	/* W */	return kv->SHIFT ? 0x57 : 0x77;
    else if (gckey == GCKeyCodeKeyE)	/* E */	return kv->SHIFT ? 0x45 : 0x65;
    else if (gckey == GCKeyCodeKeyR)	/* R */	return kv->SHIFT ? 0x52 : 0x72;
    else if (gckey == GCKeyCodeKeyT)	/* T */	return kv->SHIFT ? 0x54 : 0x74;
    else if (gckey == GCKeyCodeKeyY)	/* Y */	return kv->SHIFT ? 0x59 : 0x79;
    else if (gckey == GCKeyCodeKeyU)	/* U */	return kv->SHIFT ? 0x55 : 0x75;
    else if (gckey == GCKeyCodeKeyI)	/* I */	return kv->SHIFT ? 0x49 : 0x69;
    else if (gckey == GCKeyCodeKeyO)	/* O */	return kv->SHIFT ? 0x4F : 0x6F;
    else if (gckey == GCKeyCodeKeyP)	/* P */	return kv->SHIFT ? 0x50 : 0x70;

    else if (gckey == GCKeyCodeKeyA)	/* A */	return kv->SHIFT ? 0x41 : 0x61;
    else if (gckey == GCKeyCodeKeyS)	/* S */	return kv->SHIFT ? 0x53 : 0x73;
    else if (gckey == GCKeyCodeKeyD)	/* D */	return kv->SHIFT ? 0x44 : 0x64;
    else if (gckey == GCKeyCodeKeyF)	/* F */	return kv->SHIFT ? 0x46 : 0x66;
    else if (gckey == GCKeyCodeKeyG)	/* G */	return kv->SHIFT ? 0x47 : 0x67;
    else if (gckey == GCKeyCodeKeyF)	/* H */	return kv->SHIFT ? 0x48 : 0x68;
    else if (gckey == GCKeyCodeKeyJ)	/* J */	return kv->SHIFT ? 0x4A : 0x6A;
    else if (gckey == GCKeyCodeKeyK)	/* K */	return kv->SHIFT ? 0x4B : 0x6B;
    else if (gckey == GCKeyCodeKeyL)	/* L */	return kv->SHIFT ? 0x4C : 0x6C;

    else if (gckey == GCKeyCodeKeyZ)	/* Z */	return kv->SHIFT ? 0x5A : 0x7A;
    else if (gckey == GCKeyCodeKeyX)	/* X */	return kv->SHIFT ? 0x58 : 0x78;
    else if (gckey == GCKeyCodeKeyC)	/* C */	return kv->SHIFT ? 0x43 : 0x63;
    else if (gckey == GCKeyCodeKeyV)	/* V */	return kv->SHIFT ? 0x56 : 0x76;
    else if (gckey == GCKeyCodeKeyB)	/* B */	return kv->SHIFT ? 0x42 : 0x62;
    else if (gckey == GCKeyCodeKeyN)	/* N */	return kv->SHIFT ? 0x4E : 0x6E;
    else if (gckey == GCKeyCodeKeyM)	/* M */	return kv->SHIFT ? 0x4D : 0x6D;

    else if (gckey == GCKeyCodeEscape) {              /*ESC*/ kv->key = VK_ESCAPE;	return 0x1B;}
    else if (gckey == GCKeyCodeDeleteOrBackspace){    /* BS*/ kv->key = VK_BACK;		return 0x08;}
    else if (gckey == GCKeyCodeDeleteOrBackspace){    /*DEL*/ kv->key = VK_DELETE;	return 0x08;}
    else if (gckey == GCKeyCodeTab){                  /*TAB*/ kv->key = VK_TAB;		return 0x09;}
    else if (gckey == GCKeyCodeReturnOrEnter){        /*RET*/ kv->key = VK_RETURN;	return 0x0D;}

    else if (gckey == GCKeyCodeKeypad0)	/* 0 */     return 0x30;
    else if (gckey == GCKeyCodeKeypad1)	/* 1 */     return 0x31;
    else if (gckey == GCKeyCodeKeypad2)	/* 2 */     return 0x32;
    else if (gckey == GCKeyCodeKeypad3)	/* 3 */     return 0x33;
    else if (gckey == GCKeyCodeKeypad4)	/* 4 */     return 0x34;
    else if (gckey == GCKeyCodeKeypad5)	/* 5 */     return 0x35;
    else if (gckey == GCKeyCodeKeypad6)	/* 6 */     return 0x36;
    else if (gckey == GCKeyCodeKeypad7)	/* 7 */     return 0x37;
    else if (gckey == GCKeyCodeKeypad8)	/* 8 */     return 0x38;
    else if (gckey == GCKeyCodeKeypad9)	/* 9 */     return 0x39;
    else if (gckey == GCKeyCodeKeypadAsterisk)        /* * */ return 0x2a;
    else if (gckey == GCKeyCodeKeypadPlus)            /* + */ return 0x2b;
    else if (gckey == GCKeyCodeKeypadEqualSign)       /* = */ return 0x3d;
    else if (gckey == GCKeyCodeKeypadPeriod)          /* . */ return 0x2e;

    else if (gckey == GCKeyCodeHyphen)              /* = -   */	return kv->SHIFT ? 0x3D : 0x2D;
    else if (gckey == GCKeyCodeGraveAccentAndTilde) /* ^ ~   */	return kv->SHIFT ? 0x7E : 0x5E;
    else if (gckey == GCKeyCodeOpenBracket)         /* @ `   */	return kv->SHIFT ? 0x60 : 0x40;
    else if (gckey == GCKeyCodeCloseBracket)	    /* [ {   */	return kv->SHIFT ? 0x7B : 0x5B;
    else if (gckey == GCKeyCodeSemicolon)           /* + ;   */	return kv->SHIFT ? 0x2B : 0x3B;
    else if (gckey == GCKeyCodeQuote)               /* * :   */	return kv->SHIFT ? 0x2A : 0x3A;
    else if (gckey == GCKeyCodeCloseBracket)        /* ] }   */	return kv->SHIFT ? 0x7D : 0x5D;
    else if (gckey == GCKeyCodeComma)               /* < ,   */	return kv->SHIFT ? 0x3C : 0x2C;
    else if (gckey == GCKeyCodePeriod)              /* > .   */	return kv->SHIFT ? 0x3E : 0x2E;
    else if (gckey == GCKeyCodeSlash)               /* / ?   */	return kv->SHIFT ? 0x3F : 0x2F;
    else if (gckey == GCKeyCodeInternational1)      /* \ _   */	return kv->SHIFT ? 0x5F : 0x5C;
    else if (gckey == GCKeyCodeInternational3)      /* \ |   */	return kv->SHIFT ? 0x7C : 0x5C;
	else if (gckey == GCKeyCodeUpArrow)		kv->key = VK_UP;
	else if (gckey == GCKeyCodeDownArrow)	kv->key = VK_DOWN;
	else if (gckey == GCKeyCodeLeftArrow)	kv->key = VK_UP;
	else if (gckey == GCKeyCodeRightArrow)	kv->key = VK_RIGHT;
	else if (gckey == GCKeyCodeF9)			kv->key = VK_F9;
	else if (gckey == GCKeyCodeSpacebar)	kv->key = VK_SPACE;

    return '\0';
}
//static void utf8_to_sjis(const char *target, char *buff, int siz) {
//    NSString *utf8 = [[NSString alloc] initWithUTF8String:(void *)target];
//    NSData *sjis = [utf8 dataUsingEncoding:NSShiftJISStringEncoding];
//    memset(buff, 0, siz);
//    if (sjis.length > 0) {
//        memcpy(buff, sjis.bytes, sjis.length);
//    }
//}
static bool is_alpha(LPKEYEVT kv) {
	if (kv->ascii >= 'a' && kv->ascii <= 'z')
		return true;
	if (kv->ascii >= 'A' && kv->ascii <= 'Z')
		return true;
	return false;
}
static bool is_numeric(LPKEYEVT kv) {
	return (kv->ascii >= '0' && kv->ascii <= '9');
}
static void _toggle_print(LPCFGITEM self, char *vram, char on) {
	memset(vram, SP, 78);
	strcpy(vram, self->name);
	if (self->value == 1)
		vram[40] = on;
}
static int _controller_common_toggle(LPCFGITEM self, int verb, LPVOID param) {
	switch (verb) {
		case VERB_EVENT: {
			LPKEYEVT kv = (LPKEYEVT)param;
			if (kv->key == VK_SPACE || kv->key == VK_RETURN) {
				self->value = self->value == 1 ? 0 : 1;
				self->controller(self, VERB_COMMIT, NULL);
				return ACT_DONE;
			} else
				return ACT_NONE;
		}
		case VERB_PRINT:
		case VERB_PRINT_SEL:
			_toggle_print(self, (char*)param, 'X');
			break;
	}
	return ACT_NONE;
}
static int _controller_tenkeyless(LPCFGITEM self, int verb, LPVOID param) {
	switch (verb) {
		case VERB_INIT:
			self->value = n80lx.tenkeyless ? 1 : 0;
			break;
		case VERB_COMMIT:
			n80lx.tenkeyless = self->value == 1;
			break;
		default:
			return _controller_common_toggle(self, verb, param);
	}
	return ACT_NONE;
}
static int _controller_interrace(LPCFGITEM self, int verb, LPVOID param) {
	switch (verb) {
		case VERB_INIT:
			self->value = n80device.param.interlace ? 1 : 0;
			break;
		case VERB_COMMIT:
			n80device.param.interlace = self->value == 1;
			n80device.blt.chg_palette(n80device.param.interlace);
			break;
		default:
			return _controller_common_toggle(self, verb, param);
	}
	return ACT_NONE;
}
static int _controller_nowait(LPCFGITEM self, int verb, LPVOID param) {
	switch (verb) {
		case VERB_INIT:
			self->value = n80lx.nowait ? 1 : 0;
			break;
		case VERB_COMMIT:
			n80lx.nowait = self->value == 1;
			break;
		default:
			return _controller_common_toggle(self, verb, param);
	}
	return ACT_NONE;
}
static int _controller_common_filename(LPCFGITEM self, int verb, LPVOID param, char *buff, const char *defval) {
	switch (verb) {
		case VERB_INIT:
			break;
		case VERB_EVENT: {
			LPKEYEVT kv = (LPKEYEVT)param;
			if (kv->key == VK_RETURN) {
				self->controller(self, VERB_COMMIT, kv);
				return ACT_DONE;
			}
			int len = (int)strlen(buff);
			if (len > self->value)
				return ACT_NONE;
			if (kv->key == VK_BACK && len > 0) {    // [BS] Key pressed
				buff[len - 1] = '\0';
				return ACT_DONE;
			} else if (len < self->value && self->controller(self, VERB_TEXTIN_CH, kv) == CHAR_VALID) {
				buff[len] = kv->ascii;
				buff[len+1] = '\0';
				return ACT_DONE;
			}
			return ACT_NONE;
		}
		case VERB_PRINT:
		case VERB_PRINT_SEL: {
			//_textin_print(self, (char*)param, buff, self->value, verb == VERB_PRINT_SEL);
			//break;
			char *vram = (char*)param;
			int l = (int)strlen(buff);
			memset(vram, SP, 78);
			strcpy(vram, self->name);
			for (int i = 0; i <  self->value; ++i) {
				char ch = *buff++;
				if (ch == '\0')
					break;
				vram[40 + i] = ch;
			}
			// show guide cursor when selected
			if (verb == VERB_PRINT_SEL) {
				if (l <  self->value)
					vram[40 + l] = 0x1D;
			}
		}
		case VERB_TEXTIN_CH: {
			LPKEYEVT kv = (LPKEYEVT)param;
			if (is_alpha(kv))
				return CHAR_VALID;
			if (is_numeric(kv))
				return CHAR_VALID;
			if (kv->ascii == '_' || kv->ascii == '.')
				return CHAR_VALID;
			if (kv->ascii == '-' && strlen(buff) > 0)
				return CHAR_VALID;
			return CHAR_INVALID;
		}
		case VERB_COMMIT:
			if (strlen(buff) == 0 && defval != NULL)
				strcpy(buff, defval);
			break;
	}
	return ACT_NONE;
}
static int _controller_cmtfile(LPCFGITEM self, int verb, LPVOID param) {
	return _controller_common_filename(self, verb, param, n80lx.save_cmt_name, "casette.cmt");
}
static int _controller_n80rom(LPCFGITEM self, int verb, LPVOID param) {
	return _controller_common_filename(self, verb, param, n80lx.n80rom_filename, "PC-8001.ROM");
}
static int _controller_fdurom(LPCFGITEM self, int verb, LPVOID param) {
	return _controller_common_filename(self, verb, param, n80lx.fdurom_filename, NULL);
}
static int _controller_image_slot(LPCFGITEM self, int verb, LPVOID param) {
	switch (verb) {
		case VERB_INIT:
			self->value = self->arg == NULL ? 0 : 1;
			break;
		case VERB_EVENT: {
			LPKEYEVT kv = (LPKEYEVT)param;
			if (self->value == 1) {
				if (kv->key == VK_SPACE || kv->key == VK_RETURN) {
					self->parent->jmp_arg  = self->arg;
					self->parent->jmp_page = 1;
					return ACT_CHANGEPAGE;
				} else if (kv->key == VK_DELETE) {
					int slot_no = (self->id & 0x0f) - 1;
					n80device.fdu.mgr.detach(&n80device.fdu.mgr, slot_no);
					self->arg   = NULL;
					self->value = 0;
					return ACT_DONE;
				}
			} else if (kv->key == VK_INSERT || kv->key == VK_F9) {
				int slot_no = (self->id & 0x0f) - 1;
				LPVOID	ptr = self->parent->parent->parent;
                n80config*	n80cfg = (__bridge n80config*)ptr;
				[n80cfg chooser:slot_no];

				self->parent->jmp_arg  = NULL;
				self->parent->jmp_page = 0;
				return ACT_CHANGEPAGE;
			}
			return ACT_NONE;
		}
		case VERB_PRINT:
		case VERB_PRINT_SEL:
			_toggle_print(self, (char*)param, '\x1c');
			break;
		case VERB_COMMIT:
			break;
	}
	return ACT_NONE;
}
static int _controller_drive_select(LPCFGITEM self, int verb, LPVOID param) {
	switch (verb) {
		case VERB_INIT:
			self->value = ((LPFLOPPYDISK)self->arg)->unit_no;
			break;
		case VERB_EVENT: {
			LPKEYEVT kv = (LPKEYEVT)param;
			if (kv->key == VK_SPACE || kv->key == VK_RETURN) {
				self->controller(self, VERB_COMMIT, NULL);
				return ACT_CHANGEPAGE;
			} else if (kv->key == VK_LEFT) {
				switch (self->value) {
					case 0:	self->value = -1;	break;
					case 1: self->value = 0;	break;
					default:self->value = 1;	break;
				}
				return ACT_DONE;
			} else if (kv->key == VK_RIGHT) {
				switch (self->value) {
					case 0:	self->value = 1;	break;
					case 1: self->value = -1;	break;
					default:self->value = 0;	break;
				}
				return ACT_DONE;
			}
			return ACT_NONE;
		}
		case VERB_PRINT:
		case VERB_PRINT_SEL: {
			LPFLOPPYDISK disk = (LPFLOPPYDISK) self->arg;
			char *vram = (char*)param;
			memset(vram, SP, 78);
			strcpy(vram, self->name);
			// copy disk name
			char *title = disk->header->title[0] == 0 ? "{NO TITLE}\0\0\0\0\0\0\0" : disk->header->title;
			for (int j = 0; j < 17; ++j) {
				char c = title[j];
				vram[j + 4] = c == '\0' ? '.' : c;
			}
			// mount status
			char c = ' ';
			switch (self->value) {
				case 0: c = 'A'; break;
				case 1: c = 'B'; break;
			}
			vram[40] = c;
			break;
		}
		case VERB_COMMIT: {
			LPFLOPPYDISK disk = (LPFLOPPYDISK) self->arg;
			if (disk->unit_no != self->value) {
				if (disk->unit_no == 0 || disk->unit_no == 1)
					n80device.fdu.fdd.eject(&n80device.fdu.fdd, disk->unit_no);
				if (self->value == 0 || self->value ==  1) {
					n80device.fdu.fdd.eject(&n80device.fdu.fdd, self->value);
					n80device.fdu.fdd.insert(&n80device.fdu.fdd, self->value, disk);
				}
				self->value = disk->unit_no;
				self->parent->reflesh(self->parent);
			}
			break;
		}
	}
	return ACT_NONE;
}
/**--------------------------------------------------------------------------
 **	page methods
 **------------------------------------------------------------------------*/
static void drawitem(LPCFGITEM item, int line, BOOL selected) {
	int row = line + 9;
	item->controller(item, selected ? VERB_PRINT_SEL : VERB_PRINT, &item->parent->parent->text[row][1]);
	for (int i = 1; i < 79; ++i)
		item->parent->parent->attr[row][i] = selected ? 0x22 : 0x07;
	if (selected) {
		memset(&item->parent->parent->text[7][1], SP, 78);
		sprintf(&item->parent->parent->text[7][1], " hint: %s", item->hint);
	}
}

static LPCFGITEM get_item(LPCFGPAGE self) {
	return &self->items[self->cur_idx];
}

static int call_current_item(LPCFGPAGE self, int verb, LPVOID arg) {
	LPCFGITEM item = self->get_item(self);
	return item->controller(item, verb, arg);
}

static void page_reflesh(LPCFGPAGE self) {
	for (int i = 0; i < self->item_cnt; ++i) {
		LPCFGITEM item = &self->items[i];
		item->controller(item, VERB_INIT, NULL);
		self->parent->drawitem(item, i, i == self->cur_idx);
	}
	self->parent->show(self->parent);
}

static void prepare_page(LPCFGPAGE self, LPVOID arg) {
	self->item_cnt = 0;
	self->cur_idx = 0;
	switch (self->page) {
		case 0:
			for (int r = 0; self->items[r].type != TYPE_TERMINATOR; ++r) {
				self->items[r].parent = self;
				self->item_cnt++;
				if (self->items[r].type == TYPE_IMGSLOT) {
					int slot = (self->items[r].id & 0x0f) - 1;
					self->items[r].arg = n80device.fdu.mgr.slots[slot];
				}
				self->items[r].controller(&self->items[r], VERB_INIT, NULL);
				self->parent->drawitem(&self->items[r], r, r == 0);
			}
			break;
		default: {
			LPDISKIMAGE image	= (LPDISKIMAGE)arg;
			for (int r = 0; r < image->disk_count; ++r) {
				LPFLOPPYDISK disk = image->disks[r];
				LPCFGITEM item = &self->items[r];
				item->parent = self;
				item->arg = disk;
				self->item_cnt++;
				self->items[r].controller(&self->items[r], VERB_INIT, NULL);
				self->parent->drawitem(item, r, r == 0);
			}
		}
	}
}

static int handle_evt_page(LPCFGPAGE self, LPKEYEVT kv) {
	LPCFGITEM item = self->get_item(self);
	int action = item->controller(item, VERB_EVENT, kv);
	switch (self->page) {
		case 0:
			if (kv->key == VK_ESCAPE) {
				self->call(self, VERB_COMMIT, NULL);
				return ACT_QUIT;
			}
			break;
		default: {
			if (action == ACT_QUIT || kv->key == VK_ESCAPE) {
				self->call(self, VERB_COMMIT, NULL);
				self->jmp_arg  = NULL;
				self->jmp_page = 0;
				return ACT_CHANGEPAGE;
			}
			break;
		}
	}
	return action;
}
/**--------------------------------------------------------------------------
 **	config methods
 **------------------------------------------------------------------------*/
static void show(LPCONFIG self) {
}

static void init_page(LPCONFIG self, int page, LPVOID arg) {
	self->page		= page;
	self->maxpages	= 2;

	// init page objects
	for (int p = 0; p < self->maxpages; ++p) {
		self->pages[p].parent	= self;
		self->pages[p].page		= p;
		self->pages[p].items	= page_items[p];
		self->pages[p].get_item	= get_item;
		self->pages[p].call		= call_current_item;
		self->pages[p].reflesh	= page_reflesh;
		self->pages[p].prepare  = prepare_page;
		self->pages[p].handler  = handle_evt_page;
	}
	// clear and draw base screen 
	for (int r = 0; r < 25; ++r) {
		for (int c = 0; c < 80; ++c) {
			BYTE atr = 0x23;
			if ((r == 4 || r == 5) && (c > 0 && c < 79))
				atr = 0x27;
			else if (r == 7 && (c > 0 && c < 79))
				atr = 0x26;
			else if ((r > 8 && r < 24) && (c > 0 && c < 79))
				atr = 0x07;
			self->attr[r][c] = atr;
			self->text[r][c] = SP;
		}
	}
    const char *title1 = "    \xea\xeaN80\xea\xea PC-8001 Emulator for macOS (14.0[Sonoma] or later)";
	const char *title2 = "    Copyright 1993-2021 (C) H.Kanemoto, All Right Reserved.";
	const char *title3 = "System Config";
	const char *title4 = "usage:  [ESC] Quit/Back, [\x1e\x1f] Move, [SP] Toggle, [\x1d\x1c] Value";
	const char *title5 = "hint: ";
	memcpy(&self->text[1][2], title1, strlen(title1));
	memcpy(&self->text[2][2], title2, strlen(title2));
	memcpy(&self->text[4][2], title3, strlen(title3));
	memcpy(&self->text[5][2], title4, strlen(title4));
	memcpy(&self->text[7][2], title5, strlen(title5));

	self->pages[page].prepare(&self->pages[page], arg);
}

static BOOL dispatch(LPCONFIG self, LPKEYEVT kv) {
	LPCFGPAGE page = &self->pages[self->page];

	int action = page->handler(page, kv);
	if (action == ACT_DONE) {
		self->drawitem(page->get_item(page), page->cur_idx, true);
		self->show(self);
	} else if (action == ACT_CHANGEPAGE) {
		self->init_page(self, page->jmp_page, page->jmp_arg);
		self->show(self);
	} else if (action == ACT_QUIT) {
		return true;
	}
	else if ((kv->key == VK_TAB && !kv->SHIFT) || kv->key == VK_DOWN) { // [TAB] move to next item
		if (page->item_cnt > 1) {
			page->call(page, VERB_COMMIT, NULL);
			self->drawitem(page->get_item(page), page->cur_idx, false);
			if (++page->cur_idx >= page->item_cnt)
				page->cur_idx = 0;
			self->drawitem(page->get_item(page), page->cur_idx, true);
			self->show(self);
		}
	} else if ((kv->key == VK_TAB && kv->SHIFT) || kv->key == VK_UP) {
		if (page->item_cnt > 1) {
			page->call(page, VERB_COMMIT, NULL);
			self->drawitem(page->get_item(page), page->cur_idx, false);
			if (--page->cur_idx < 0)
				page->cur_idx = page->item_cnt - 1;
			self->drawitem(page->get_item(page), page->cur_idx, true);
			self->show(self);
		}
	}
	return false;
}

static void config(LPCONFIG self) {
	self->init_page(self, 0, NULL);
	self->show(self);
}
/**--------------------------------------------------------------------------
 **		Initializer
 **------------------------------------------------------------------------*/
export void init_config(LPCONFIG self, LPVOID parent) {
	self->init		= init_config;
	self->open      = config;
	self->init_page = init_page;
	self->dispatch	= dispatch;
	self->drawitem	= drawitem;
	self->show		= show;
	self->parent	= parent;
}
//-----------------------------------------------------------------------------
//    Emulator Objective-C Class
//-----------------------------------------------------------------------------
@implementation n80config

-(id)initWithN80:(n80em*)n80em config:(n80config_data*)config {
    if (self = [super init]) {
        self->n80em = n80em;
        self->data  = config;
        n80device.param.interlace       = config.interlace;
        n80device.param.use_new_pcg     = config.chiqlappe;
        n80device.vrtc.crt_hires_mode   = config.hires_crt ? _CRT_HIRES : _CRT_NORMAL;
		n80lx.nowait					= config.nowait;
		n80lx.tenkeyless				= config.tenkeyless;
	    n80lx.n80rom_filename			= &self->data->slot[0][0];
	    n80lx._4throm_filename			= &self->data->slot[1][0];
	    n80lx.fdurom_filename			= &self->data->slot[2][0];
    }
    return self;
}

-(n80config const *)getParam {
    return (n80config const *)self->data;
}

-(void)enter_config {
    n80lx.tmp_save_screen_mode   = n80device.crtc._screen_mode;
    n80lx.tmp_save_text_visible  = n80device.crtc._text_visible;
    n80lx.tmp_save_use_new_pcg   = n80device.param.use_new_pcg;

    [self->n80em.n80gr chg_bltfunc:3];
    n80device.crtc._text_visible = 0;	// text visible
    n80device.param.use_new_pcg = false;// disable new pcg

    init_config(&n80lx.config, (__bridge LPVOID)(self));
	n80lx.config.open(&n80lx.config);

	[self->n80em.n80gr draw_config:n80lx.config.text attr:n80lx.config.attr];
}

-(void)leave_config {
    n80device.crtc._screen_mode  = n80lx.tmp_save_screen_mode;
    n80device.crtc._text_visible = n80lx.tmp_save_text_visible;
    n80device.param.use_new_pcg  = n80lx.tmp_save_use_new_pcg;

    self->data.interlace  = n80device.param.interlace;
    self->data.chiqlappe  = n80device.param.use_new_pcg;
    self->data.hires_crt  = n80device.vrtc.crt_hires_mode == _CRT_HIRES;
    self->data.tenkeyless = n80lx.tenkeyless;
    self->data.nowait     = n80lx.nowait;

    [self->n80em.n80gr chg_bltfunc:n80lx.tmp_save_screen_mode];
    [self reflect_menu_state];
}

-(void)reflect_menu_state {
    AppDelegate *app = [[NSApplication sharedApplication] delegate];
    if (n80device.param.interlace)
        [[app menuInterlace] setState:NSControlStateValueOn];
    else
        [[app menuInterlace] setState:NSControlStateValueOff];
    if (n80lx.tenkeyless)
        [[app menuTenKeyLess] setState:NSControlStateValueOn];
    else
        [[app menuTenKeyLess] setState:NSControlStateValueOff];
    if (n80device.pcg_mode != 0)
        [[app menuPCG] setState:NSControlStateValueOn];
    else
        [[app menuPCG] setState:NSControlStateValueOff];
    if (n80device.param.use_new_pcg)
        [[app menuNewPCG] setState:NSControlStateValueOn];
    else
        [[app menuNewPCG] setState:NSControlStateValueOff];
	if (n80device.fdu.fdd.drives[0].media_inserted)
        [[app menuDrive1] setState:NSControlStateValueOn];
	else
        [[app menuDrive1] setState:NSControlStateValueOff];
	if (n80device.fdu.fdd.drives[1].media_inserted)
        [[app menuDrive2] setState:NSControlStateValueOn];
	else
        [[app menuDrive2] setState:NSControlStateValueOff];
}

-(bool)OnKeyDown:(BYTE)key code:(GCKeyCode)c {
	bool result = false;
    n80lx.config.kv.code = key;
    n80lx.config.kv.ascii = to_ascii(&n80lx.config.kv, c);
    printf("%c", n80lx.config.kv.ascii);
    if (c == GCKeyCodeLeftShift || c == GCKeyCodeRightShift)
        n80lx.config.kv.SHIFT = true;
    else
		result = n80lx.config.dispatch(&n80lx.config, &n80lx.config.kv);

	[self->n80em.n80gr draw_config:n80lx.config.text attr:n80lx.config.attr];
	return result;
}

-(void)OnKeyUp:(BYTE)key code:(GCKeyCode)c {
    if (c == GCKeyCodeLeftShift || c == GCKeyCodeRightShift) {
        n80lx.config.kv.SHIFT = false;
    } else {
	    n80lx.config.kv.code = key;
        to_ascii(&n80lx.config.kv, c);
    }
    return;
}

-(void)chooser:(int)slot_no {
	[self->n80em req_open_d88_file_core:slot_no];
	[self reflect_menu_state];
}

@end
//-----------------------------------------------------------------------------
//    Emulator Objective-C Class
//-----------------------------------------------------------------------------
@implementation n80config_data
-(id)init {
    if (self = [super init]) {
        memset(self->slot, 0, sizeof(self->slot));
    }
    return self;
}
-(void)set_char:(char)c index:(int)index slot:(int)slot {
    self->slot[slot][index] = (char)c;
}
-(char)get_char:(int)index slot:(int)slot {
    return self->slot[slot][index];
}
-(int)get_size:(int)slot {
    return (int)strlen(&self->slot[slot][0]);
}

@end
