//
//  n80em.m
//
//  Copyright (C) 1993-2019 Hiroshi Kanemoto.
//    Released under the MIT license
//    http://opensource.org/licenses/mit-license.php
//
//     Created:    2019/05/12    H.Kanemoto
//
#import "n80sound.h"
#import <stdatomic.h>
#include <limits.h>
#include <mach/mach_time.h>
#include <AudioUnit/AudioUnit.h>
extern N80LX        n80lx;
//-----------------------------------------------------------------------------
//    Emulator C Functions
//-----------------------------------------------------------------------------
#define SAMPLE_RATE (96000.0)

static double       state_cnt_per_frame = 0;    // 3993600Hz / sampling_rate { 41.6T(96kHz)/83.2T(48kHz)/96.23T(41.5kHz) }
static WAVETsCONV   beep_wave;
static SAMPLETsBANK beep_bank;
static SAMPLETsBANK sound_bank[3];
static SAMPLETsARR  beep_copy_arr;
static SAMPLETsARR  sound_copy_arr[3];
//-----------------------------------------------------------------------------
//    one-wave frame/T function
//-----------------------------------------------------------------------------
static void calc_wave_frame(LPWAVETsCONV wave, int sampling_rate, int tone_hz) {
    if (tone_hz == 0) {     // lo-tone min (silent)
        wave->tone = tone_hz;
        wave->T = 0;
        wave->halfT = 0;
        wave->frames = 0;
        wave->halfs[0] = 0;
        wave->halfs[1] = 0;
    } else {
        wave->tone = tone_hz;
        wave->T = n80device.param.hz / tone_hz;
        wave->halfT = wave->T / 2;
        wave->frames = sampling_rate / tone_hz; // ex) 96,000Hz / 2400Hz = 40 frames
        if (wave->frames < 2) {
            wave->frames = 2;   // hi-tone max
        }
        wave->halfs[0] = wave->frames / 2;
        wave->halfs[1] = wave->halfs[0];
        if ((wave->frames % 2) > 0)
            ++wave->halfs[1];
        int t = wave->T / wave->frames;
        int a = wave->T % wave->frames;
        for (int i = 0; i < wave->frames; ++i) {
            wave->arr[i] = t;
        }
        if (a > 0) {
            int odd = 0, b = 0;
            for (int i = 0; i < a; ++i) {
                if (b < wave->frames) {
                    ++wave->arr[b];
                } else {
                    odd = odd == 0 ? 1 : 0;
                    b = odd;
                    ++wave->arr[b];
                }
                b += 2;
            }
        }
    }
    wave->offset = 0;
    fprintf(stderr, "%dHz tone one-wave frames = %d (sampling rate %dHz)\n", tone_hz, wave->frames, sampling_rate);
}
static DWORD iterator_bank(LPSAMPLETsARR arr) {
    for (int i = arr->index; i < arr->count; ++i) {
        if (arr->arr[i].ticks > 0) {
            --arr->arr[i].ticks;
            return arr->arr[i].state;
        }
    }
    return 0;
}
static void beep_pcm_maker(LPSAMPLETsARR arr, int au_frames, float *buffer, LPWAVETsCONV wave) {
    for (int i = 0; i < au_frames; ++i) {
        float volume = 0;
        if (wave->offset >= wave->frames) {
            wave->offset = 0;
        }
        for (int t = 0; t < wave->arr[wave->offset]; ++t) {
            DWORD on_off = iterator_bank(arr);
            if (on_off != 0)
                volume += 1.0f;
            else
                volume -= 1.0f;
        }
        if (wave->offset < wave->halfs[0])  // lo-wave always silent
            buffer[i] = -1;
        else {
            volume /= (float)wave->arr[wave->offset];
            buffer[i] = (short)volume;
        }
        ++wave->offset;
    }
}
//-----------------------------------------------------------------------------
//    supplement function
//-----------------------------------------------------------------------------
static void append_tick_to_array(LPSAMPLETsBANK bank, DWORD new_state) {
    // 現在のTick累積を配列に追加する
    LPSAMPLETsARR arr = &bank->banks[bank->bank];
    if (arr->count < MAXSTsARRAYS - 1) {
        arr->arr[arr->count].ticks   = bank->current.ticks;
        arr->arr[arr->count++].state = bank->current.state;
        arr->ticks_sum += bank->current.ticks;
    } else
        fprintf(stderr, "beep ticks array overflow\n");
    
    bank->current.ticks = 0;
    bank->current.state = new_state;
}
static void advance_tick_to_array(LPSAMPLETsBANK bank, DWORD diff) {
    DWORD add = 0;
    if (bank->current.ticks > diff) {
        bank->current.ticks -= diff;
        add = diff;
    } else {
        add = bank->current.ticks;
        bank->current.ticks = 0;
    }
    if (add > 0) {
        LPSAMPLETsARR arr = &bank->banks[bank->bank];
        if (arr->count < MAXSTsARRAYS - 1) {
            arr->arr[arr->count].ticks = add;
            arr->arr[arr->count].state = bank->current.state;
            arr->ticks_sum += add;
            ++arr->count;
        } else if (arr->arr[MAXSTsARRAYS - 1].state == bank->current.state) {
            arr->arr[MAXSTsARRAYS - 1].ticks += add;
            arr->ticks_sum += add;
        } else
            fprintf(stderr, "advanceTick: bank overrun\n");
    }
}
//-----------------------------------------------------------------------------
//    AudioUnit Callback function
//-----------------------------------------------------------------------------
static OSStatus    audioCallback(
    void                           *inRefCon,
    AudioUnitRenderActionFlags     *ioActionFlags,
    const AudioTimeStamp           *inTimeStamp,
    UInt32                         inBusNumber,
    UInt32                         inNumberFrames,
    AudioBufferList                *ioData) {
    memset(ioData->mBuffers[0].mData, 0, ioData->mBuffers[0].mDataByteSize);
    n80sound *n80snd = (__bridge n80sound*)inRefCon;
    [n80snd callback:inNumberFrames ioData:ioData];

    return noErr;
}
//-----------------------------------------------------------------------------
//    Emulator Objective-C Class
//-----------------------------------------------------------------------------
@implementation n80sound

-(id)initWithN80:(n80em*)n80em {
    if (self = [super init]) {
        self->n80em = n80em;
        // init member variables
        self->gOutputUnit = NULL;
        // lock for n80 sync
        self->sync = [[NSRecursiveLock alloc] init];

        // beep/sound bank init
        memset(&beep_bank, 0, sizeof(beep_bank));
        memset(&sound_bank, 0, sizeof(sound_bank));

        if (![self prepareAudio])
            return nil;
    }
    return self;
}

-(void)dealloc {
    NSLog(@"n80sound dealloc");
    [self reqDispose];
}

-(N80LX const *)getN80 {
    return &n80lx;
}

-(void)lock {
    [self->sync lock];
}

-(void)unlock {
    [self->sync unlock];
}

-(void)suspend {
    if (gOutputUnit != NULL) {
        AudioOutputUnitStop(gOutputUnit);
        AudioUnitUninitialize(gOutputUnit);
        AudioComponentInstanceDispose(gOutputUnit);
        self->gOutputUnit = NULL;
    }
}

-(BOOL)resume {
    return [self prepareAudio];
}

- (void)reqDispose {
    // stop sound unit
    if (gOutputUnit != NULL) {
        AudioOutputUnitStop(gOutputUnit);
        AudioUnitUninitialize(gOutputUnit);
        AudioComponentInstanceDispose(gOutputUnit);
    }
    self->sync = NULL;
}
-(void)append_sound_ticks:(DWORD)T {
    // CPU STATE数をALSA frame数に変換して累積する
    double frames = n80lx.frame_keisu * T;
    [self lock];
    beep_bank.current.ticks     += T;
    sound_bank[0].current.ticks += frames;
    sound_bank[1].current.ticks += frames;
    sound_bank[2].current.ticks += frames;
    [self unlock];
}
//-----------------------------------------------------------------------------
//    Initialize Emulator
//-----------------------------------------------------------------------------
- (BOOL)prepareAudio {
    OSStatus    err = noErr;
    int         sampleRate = SAMPLE_RATE;
    
    // Open the default output unit
    AudioComponentDescription desc;
    desc.componentType = kAudioUnitType_Output;
#if TARGET_OS_OSX
    desc.componentSubType = kAudioUnitSubType_DefaultOutput;
#else
	desc.componentSubType = kAudioUnitSubType_RemoteIO;
#endif
    desc.componentManufacturer = kAudioUnitManufacturer_Apple;
    desc.componentFlags = 0;
    desc.componentFlagsMask = 0;
    
    AudioComponent comp = AudioComponentFindNext(NULL, &desc);
    if (comp == NULL) {
        NSLog(@"AudioComponentFindNext error");
        return false;
    }
    err = AudioComponentInstanceNew(comp, &gOutputUnit);
    if (comp == NULL) {
        //NSLog(@"AudioComponentInstanceNew=%d", err);
        return false;
    }
    err = AudioUnitInitialize(gOutputUnit);
    
    // Set up a callback function to generate output to the output unit
    AURenderCallbackStruct callback;
    callback.inputProc = audioCallback;
    callback.inputProcRefCon = (__bridge void*)self;
    
    err = AudioUnitSetProperty (gOutputUnit,
                                kAudioUnitProperty_SetRenderCallback,
                                //kAudioOutputUnitProperty_EnableIO,
                                kAudioUnitScope_Input,      // initial value is "kAudioUnitScope_Input". is it ok?
                                0,
                                &callback,
                                sizeof(callback));
    
    AudioStreamBasicDescription streamFormat;
    streamFormat.mSampleRate = sampleRate;                  // the sample rate of the audio stream
    streamFormat.mFormatID = kAudioFormatLinearPCM;         // the specific encoding type of audio stream
#ifdef __BIG_ENDIAN__
    streamFormat.mFormatFlags =  kLinearPCMFormatFlagIsSignedInteger | kLinearPCMFormatFlagIsBigEndian
    | kLinearPCMFormatFlagIsPacked | kAudioFormatFlagIsNonInterleaved;        //    16bit int
#else
    streamFormat.mFormatFlags =  kLinearPCMFormatFlagIsSignedInteger
    | kLinearPCMFormatFlagIsPacked | kAudioFormatFlagIsNonInterleaved;        //    16bit int
#endif
    streamFormat.mBytesPerPacket = 2;
    streamFormat.mFramesPerPacket = 1;
    streamFormat.mBytesPerFrame = 2;
    streamFormat.mChannelsPerFrame = 1;
    streamFormat.mBitsPerChannel = 16;
    
    err = AudioUnitSetProperty (gOutputUnit,
                                kAudioUnitProperty_StreamFormat,
                                kAudioUnitScope_Input,
                                0,
                                &streamFormat,
                                sizeof(AudioStreamBasicDescription));
        
    Float64 outSampleRate;
    UInt32 size = sizeof(Float64);
    
    err = AudioUnitGetProperty (gOutputUnit,
                                kAudioUnitProperty_SampleRate,
                                kAudioUnitScope_Input,
                                0,
                                &outSampleRate,
                                &size);
    // CPU Ticks変換係数: frames = (96kHz/3993600Hz*Ts)
    n80lx.sampling_rate = (int)outSampleRate;
    n80lx.frame_keisu = (double) n80lx.sampling_rate / (double) n80device.param.hz;
    state_cnt_per_frame = (double)n80device.param.hz / (double)n80lx.sampling_rate;
    fprintf(stderr, "%g T/frame\n", state_cnt_per_frame);
    calc_wave_frame(&beep_wave, n80lx.sampling_rate, 2400);

    err = AudioOutputUnitStart (gOutputUnit);
    
    return true;
}


-(void)cpu_reset {
    //[syncSound lock];                    // reset sound tick buffer
    [self lock];
    beep_bank.bank = 0;
    beep_bank.banks[0].count = 0;
    beep_bank.banks[0].ticks_sum = 0;
    beep_bank.banks[1].count = 0;
    beep_bank.banks[1].ticks_sum = 0;
    beep_bank.current.state = 0;
    beep_bank.current.ticks = 0;
    for (int i = 0; i < 3; ++i) {
        sound_bank[i].bank = 0;
        sound_bank[i].banks[0].count = 0;
        sound_bank[i].banks[0].ticks_sum = 0;
        sound_bank[i].banks[1].count = 0;
        sound_bank[i].banks[1].ticks_sum = 0;
        sound_bank[i].current.state = 0;
        sound_bank[i].current.ticks = 0;
    }
    //[syncSound unlock];
    [self unlock];
}

-(void)beep_on {
    [self lock];
    if (beep_bank.current.state == 0) {
        append_tick_to_array(&beep_bank, 1);
    }
    [self unlock];
}

-(void)beep_off {
    [self lock];
    if (beep_bank.current.state == 1) {
        append_tick_to_array(&beep_bank, 0);
    }
    [self unlock];
}

-(void)sound_on:(USHORT)freq ch:(int)ch {
    [self lock];
    if (sound_bank[ch].current.state != freq)
        for (int i = 0; i < 3; ++i) {
            USHORT f = ch == i ? freq : sound_bank[i].current.state;
            append_tick_to_array(&sound_bank[i], f);
        }
    [self unlock];
}

-(void)sound_off {
    [self lock];
    append_tick_to_array(&sound_bank[0], 0);
    append_tick_to_array(&sound_bank[1], 0);
    append_tick_to_array(&sound_bank[2], 0);
    [self unlock];
}

static inline void sample_overlay(int16_t *pcm, int val) {
    int32_t v = *pcm;
    v += val;
    if (v > SHRT_MAX)
        v = SHRT_MAX;
    else if (v < SHRT_MIN)
        v = SHRT_MIN;
    *pcm++ = (int16_t)v;
}

static inline void addTk(LPSAMPLETsARR arr, double ticks, DWORD state) {
    arr->arr[arr->count].ticks = ticks;
    arr->arr[arr->count].state = state;
    ++arr->count;
    arr->ticks_sum += ticks;
}
static inline void appendTk(LPSAMPLETsARR arr, double ticks, DWORD state) {
    if (arr->count > 0 && arr->arr[arr->count - 1].state == state)
        arr->arr[arr->count - 1].ticks += ticks;
    else {
        arr->arr[arr->count].ticks = ticks;
        arr->arr[arr->count].state = state;
        ++arr->count;
    }
    arr->ticks_sum += ticks;
}

static void exchange_bank(DWORD au_frames, int *_beep_bank, int *_sound_bank) {
    int beep_primary_bank, sound_primary_bank;

    beep_primary_bank   = beep_bank.bank;
    sound_primary_bank  = sound_bank[0].bank;
    int beep_sub_bank   = beep_primary_bank == 0 ? 1 : 0;
    int sound_sub_bank  = sound_primary_bank == 0 ? 1 : 0;
    // clear sub bank
    beep_bank.banks[beep_sub_bank].count          = 0;
    beep_bank.banks[beep_sub_bank].ticks_sum      = 0;
    beep_bank.banks[beep_sub_bank].index          = 0;
    sound_bank[0].banks[sound_sub_bank].count     = 0;
    sound_bank[0].banks[sound_sub_bank].ticks_sum = 0;
    sound_bank[1].banks[sound_sub_bank].count     = 0;
    sound_bank[1].banks[sound_sub_bank].ticks_sum = 0;
    sound_bank[2].banks[sound_sub_bank].count     = 0;
    sound_bank[2].banks[sound_sub_bank].ticks_sum = 0;

    static double total_rcv_ticks = 0;
    static double total_pick_ticks = 0;
    static double total_rollover_ticks = 0;
    static double total_frame_Ts = 0;
    static double discard_frames = 0;
    static double loop_count = 0;
    //---- Beep ---------------------------------------------------------------
    LPSAMPLETsARR arr1 = &beep_bank.banks[beep_primary_bank];
    LPSAMPLETsARR arr2 = &beep_bank.banks[beep_sub_bank];
    int frame_Ts = au_frames * state_cnt_per_frame;

    // rack of frames
    if (arr1->ticks_sum < frame_Ts) {
        DWORD diff = frame_Ts - arr1->ticks_sum;
        advance_tick_to_array(&beep_bank, diff);
    }
    //pick up need frames
    if (arr1->ticks_sum > frame_Ts) {
        LPSAMPLETsARR _arr = &beep_copy_arr;
        _arr->ticks_sum = 0;
        _arr->count     = 0;
        _arr->index     = 0;
        bool overflow = false;
        for (int i = 0; i < arr1->count; ++i) {
            DWORD ticks = arr1->arr[i].ticks;
            DWORD state = arr1->arr[i].state;
            if (!overflow) {
                DWORD T = _arr->ticks_sum + ticks;
                if (T > frame_Ts) {
                    DWORD over = T - frame_Ts;
                    ticks -= over;

                    appendTk(arr2, over, state);
                    overflow = true;
                }
                appendTk(_arr, ticks, state);
            } else {
                appendTk(arr2, ticks, state);
            }
        }
        beep_primary_bank = -1;
    }
    //---- summary ----------------------------------------------------------------
    loop_count += 1;
    double period_ticks = n80lx.sampling_rate * state_cnt_per_frame;
    total_frame_Ts += frame_Ts;
    total_rcv_ticks += arr1->ticks_sum;
    total_pick_ticks += beep_primary_bank == -1 ? beep_copy_arr.ticks_sum : arr1->ticks_sum;
    total_rollover_ticks += arr2->ticks_sum;
    if (total_frame_Ts > period_ticks/*SAMPLE_RATE*2*/) {
        //fprintf(stderr, "%g/%g\n", total_frame_Ts, period_ticks);
        //double percent = beep_frames / (total_frames) * 100;
        double percent1 = (total_rcv_ticks + total_rollover_ticks) / total_frame_Ts * 100;
        double percent2 = total_rcv_ticks / total_frame_Ts * 100;
        double percent3 = total_rollover_ticks / total_frame_Ts * 100;
        int ave_frames = (int)(total_frame_Ts / loop_count);
        double datas = (total_rcv_ticks + total_rollover_ticks) / loop_count;
        double use = (total_pick_ticks) / loop_count;
        printf("sound ave=%d rcvdata = %3.2f%%(recv:%3.2f%%, rollover=%3.2f%%) datas=%.0f(user:%.0f)\n", ave_frames, percent1, percent2, percent3, datas, use);
/*
        double ave_sync = total_frames / total_ticks * 100;
        double ave_frames = (total_frames + discard_frames) / loop_count;
        double ave_discard = discard_frames / loop_count;
        printf("sound percent = %3.2f(%3.2f)  frame = %8.1f discard = %4.2f\n", percent, ave_sync, ave_frames, ave_discard);
*/
        total_rcv_ticks = 0;
        total_pick_ticks = 0;
        total_frame_Ts = 0;
        //beep_frames = 0;
        total_rollover_ticks = 0;
        discard_frames = 0;
        loop_count = 0;
    }
    //---- PCG ----------------------------------------------------------------
    LPSAMPLETsARR arr10 = &sound_bank[0].banks[sound_primary_bank];        // current
    LPSAMPLETsARR arr20 = &sound_bank[0].banks[sound_sub_bank];            // next
    LPSAMPLETsARR arr11 = &sound_bank[1].banks[sound_primary_bank];        // current
    LPSAMPLETsARR arr21 = &sound_bank[1].banks[sound_sub_bank];            // next
    LPSAMPLETsARR arr12 = &sound_bank[2].banks[sound_primary_bank];        // current
    LPSAMPLETsARR arr22 = &sound_bank[2].banks[sound_sub_bank];            // next
    
    // rack of frames
    if (arr10->ticks_sum < au_frames) {
        double diff = au_frames - arr10->ticks_sum;
        advance_tick_to_array(&sound_bank[0], diff);
        advance_tick_to_array(&sound_bank[1], diff);
        advance_tick_to_array(&sound_bank[2], diff);
    }
    //pick up need frames
    if (arr10->ticks_sum > au_frames) {
        LPSAMPLETsARR _arr0 = &sound_copy_arr[0];
        LPSAMPLETsARR _arr1 = &sound_copy_arr[1];
        LPSAMPLETsARR _arr2 = &sound_copy_arr[2];
        _arr0->ticks_sum    = 0;
        _arr0->count        = 0;
        _arr1->ticks_sum    = 0;
        _arr1->count        = 0;
        _arr2->ticks_sum    = 0;
        _arr2->count        = 0;
        bool overflow = false;
        for (int i = 0; i < arr10->count; ++i) {
            double ticks  = arr10->arr[i].ticks;
            DWORD state0 = arr10->arr[i].state;
            DWORD state1 = arr11->arr[i].state;
            DWORD state2 = arr12->arr[i].state;
            if (!overflow) {
                double frames = _arr0->ticks_sum + ticks;
                if (frames > au_frames) {
                    double over = frames - au_frames;
                    ticks -= over;

                    addTk(arr20, over, state0);
                    addTk(arr21, over, state1);
                    addTk(arr22, over, state2);
                    overflow = true;
                }
                addTk(_arr0, ticks, state0);
                addTk(_arr1, ticks, state1);
                addTk(_arr2, ticks, state2);
            } else {
                addTk(arr20, ticks, state0);
                addTk(arr21, ticks, state1);
                addTk(arr22, ticks, state2);
            }
        }
        sound_primary_bank = -1;
    }
    // bank exchange
    beep_bank.bank = beep_sub_bank;
    for (int i = 0; i < 3; ++i) {
        sound_bank[i].bank = sound_sub_bank;
    }
    *_beep_bank  = beep_primary_bank;
    *_sound_bank = sound_primary_bank;
}

static int make_pcm(DWORD au_frames, int16_t *au_buffer, int _beep_bank, int _sound_bank) {
    int16_t *pcm = au_buffer;
    int     write_frames = 0, rollover = 0;
    float   buffer[au_frames];

    //---- Beep ---------------------------------------------------------------
    LPSAMPLETsARR arr = _beep_bank == -1 ? &beep_copy_arr : &beep_bank.banks[_beep_bank];
    beep_pcm_maker(arr, au_frames, buffer, &beep_wave);
    for (int i = 0; i < au_frames; ++i) {
        int volume = floor(WAVETOP * buffer[i]);
        pcm[i] = (short)volume;
    }
    //---- PCG ----------------------------------------------------------------
    pcm = au_buffer;
    (void)(write_frames = 0), rollover = 0;
    
    LPSAMPLETsARR   arr_0 = _sound_bank == -1 ? &sound_copy_arr[0] : &sound_bank[0].banks[_sound_bank];
    LPSAMPLETsARR   arr_1 = _sound_bank == -1 ? &sound_copy_arr[1] : &sound_bank[1].banks[_sound_bank];
    LPSAMPLETsARR   arr_2 = _sound_bank == -1 ? &sound_copy_arr[2] : &sound_bank[2].banks[_sound_bank];
    double rate = au_frames < arr_0->ticks_sum ? au_frames / arr_0->ticks_sum : 1.0;
    
    for (int i = 0; i < arr_0->count; ++i) {
        int frames = (int) ceil(arr_0->arr[i].ticks * rate);
        int remain = frames;
        int wave_frames0 = arr_0->arr[i].state > 0 ? floor((double) n80lx.sampling_rate / (double) arr_0->arr[i].state + 0.5) : 0;
        int wave_frames1 = arr_1->arr[i].state > 0 ? floor((double) n80lx.sampling_rate / (double) arr_1->arr[i].state + 0.5) : 0;
        int wave_frames2 = arr_2->arr[i].state > 0 ? floor((double) n80lx.sampling_rate / (double) arr_2->arr[i].state + 0.5) : 0;
        int half_wave_frames0 = wave_frames0 / 2;
        int half_wave_frames1 = wave_frames1 / 2;
        int half_wave_frames2 = wave_frames2 / 2;
        if (frames > 0) {
            for (int j = 0; j < frames; ++j) {
                // check if one period long over
                if (write_frames >= au_frames) {
                    break;
                }
                ++write_frames;
                --remain;
                int volume = 0;
                if (wave_frames0 > 0)    {    // ch.#0 sound on
                    if ((n80lx.phase_adjuster2 % wave_frames0) < half_wave_frames0) {
                        volume = WAVETOP;
                    } else {
                        volume = WAVEBOTTOM;
                    }
                }
                if (wave_frames1 > 0)    {    // ch.#1 sound on
                    if ((n80lx.phase_adjuster2 % wave_frames1) < half_wave_frames1) {
                        volume += WAVETOP;
                    } else {
                        volume += WAVEBOTTOM;
                    }
                }
                if (wave_frames2 > 0)    {    // ch.#2 sound on
                    if ((n80lx.phase_adjuster2 % wave_frames2) < half_wave_frames2) {
                        volume += WAVETOP;
                    } else {
                        volume += WAVEBOTTOM;
                    }
                }
                sample_overlay(pcm++, volume);
                if (++n80lx.phase_adjuster2 > 100000000)
                    n80lx.phase_adjuster2 = 0;
            }
        }
    }
    return au_frames;
}

-(void)callback:(UInt32)inNumberFrames ioData:(AudioBufferList *)ioData {
    int     beep_bank, sound_bank;
    int16_t *pcm = ioData->mBuffers[0].mData;
    DWORD   au_frames = ioData->mBuffers[0].mDataByteSize / 2;

    [self lock];
    exchange_bank(au_frames, &beep_bank, &sound_bank);
    [self unlock];
    make_pcm(au_frames, pcm, beep_bank, sound_bank);
}

@end

